# coding: utf-8
from __future__ import (absolute_import, print_function, division, unicode_literals)
from json import dumps, loads
from six.moves.xmlrpc_client import ServerProxy
import octofuss

class UnknownException(Exception):
    def __init__(self, msg):
        Exception.__init__(self, msg)


class APIKeyException(Exception):
    def __init__(self, msg):
        Exception.__init__(self, msg)


EXCEPTION_WHITELIST=set(("LookupError", "AttributeError", "ValueError", "KeyError", "APIKeyException"))

def useableBase(exc):
    if exc.__name__ in EXCEPTION_WHITELIST:
        return exc.__name__
    for b in exc.__bases__:
        ub = useableBase(b)
        if ub != None:
            return ub
    return None


class Client(octofuss.Tree):
    def __init__(self, url, apikey = ""):
        super(Client, self).__init__()
        self.url = url
        self.server = ServerProxy(url)
        self.apikey = apikey

    def _wrap_call(self, callable, *args):
        res = loads(callable(self.apikey, *args))
        if "exc" in res:
            name = str(res["exc"].get("name", "UnknownException"))
            msg = res["exc"].get("msg", "No error message provided")
            if name in EXCEPTION_WHITELIST:
                raise globals().get(name, UnknownException)(msg)
            else:
                raise UnknownException(msg)
        elif "res" in res:
            return res["res"]
        else:
            raise ValueError("value sent from server has neither an 'ext' nor a 'res' part")

    def has(self, path):
        return self._wrap_call(self.server.has, path)

    def lhas(self, path):
        return self._wrap_call(self.server.has, "/".join(path))

    def get(self, path):
        return self._wrap_call(self.server.get, path)

    def lget(self, path):
        return self._wrap_call(self.server.get, "/".join(path))

    def set(self, path, value):
        return self._wrap_call(self.server.set, path, dumps(value))

    def lset(self, path, value):
        return self._wrap_call(self.server.set, "/".join(path), dumps(value))

    def list(self, path):
        return self._wrap_call(self.server.list, path)

    def llist(self, path):
        return self._wrap_call(self.server.list, "/".join(path))

    def delete(self, path):
        return self._wrap_call(self.server.delete, path)

    def ldelete(self, path):
        return self._wrap_call(self.server.delete, "/".join(path))

    def create(self, path, value=None):
        return self._wrap_call(self.server.create, path, dumps(value))

    def lcreate(self, path, value=None):
        return self._wrap_call(self.server.create, "/".join(path), dumps(value))

    def doc(self, path):
        return self._wrap_call(self.server.doc, path)

    def ldoc(self, path):
        return self._wrap_call(self.server.doc, "/".join(path))


try:
    from twisted.web import xmlrpc

    class Server(xmlrpc.XMLRPC):
        """
        Export an octofuss.Tree over XMLRPC
        """
        def __init__(self, service, tree):
            xmlrpc.XMLRPC.__init__(self)
            self.service = service
            self.tree = tree

        def render(self, request):
            self.client = request.getClient()
            return xmlrpc.XMLRPC.render(self, request)

        def _validate_apikey(self, apikey):
            return True

        def _validate_access(self, apikey, path):
            return True

        def _wrap_call(self, apikey, denied_value, callable, *args):
            try:
                if not self._validate_apikey(apikey):
                    raise APIKeyException("Invalid API key")
                if not self._validate_access(apikey, args[0]):
                    return dumps(dict(res=denied_value))
                res = callable(*args)
                if type(res) == type(list({}.keys())):
                    res = list(res)
                return dumps(dict(res=res))

            except Exception as exc:
                import traceback
                traceback.print_exc()
                # Find a useable exception name
                name = useableBase(exc.__class__)
                if name == None:
                    name = "UnknownException"
                message = getattr(exc, "message", None)
                if message == None: message = str(exc)
                return dumps(dict(exc = dict(name = name, msg = message)))

        def xmlrpc_has(self, apikey, path):
            return self._wrap_call(apikey, False, self.tree.has, path)

        def xmlrpc_get(self, apikey, path):
            return self._wrap_call(apikey, None, self.tree.get, path)

        def xmlrpc_set(self, apikey, path, value):
            return self._wrap_call(apikey, False, self.tree.set, path, loads(value))

        def xmlrpc_list(self, apikey, path):
            return self._wrap_call(apikey, [], self.tree.list, path)

        def xmlrpc_delete(self, apikey, path):
            return self._wrap_call(apikey, False, self.tree.delete, path)

        def xmlrpc_create(self, apikey, path, value = None):
            return self._wrap_call(apikey, False, self.tree.create, path, loads(value))

        def xmlrpc_doc(self, apikey, path):
            return self._wrap_call(apikey, False, self.tree.doc, path)

except ImportError:
    # Only define XMLRPCTreeServer if we have twisted
    pass
